/** PieChartPanel.java provides a pie chart.
 *  Attributes: arrays of slice labels, slice colors, slice sizes, and
 *              slice colors; number of slices; dimensions of window frame;
 *              formatter to display numbers and percentages
 *  Methods:    constructor of a pie chart with a set of slice labels and
 *              values; paintComponent(); draw slice with a given number,
 *              start and arc angles; set slice color or value
 *              private method to compute percentages and a rounding method
 */

import java.awt.*;               // Graphics, Color, ...
import javax.swing.JPanel;
import java.text.NumberFormat;
import ann.util.Controller;

class PieChartPanel extends JPanel
{
  private final static int
    MARGIN            = 25,
    TOP_MARGIN        = 25,
    CIRCLE_DIAMETER   = 200,
    CIRCLE_LEGEND_GAP = 40,
    LEGEND_OFFSET     = TOP_MARGIN + CIRCLE_DIAMETER + CIRCLE_LEGEND_GAP,
    LEGEND_BOX_SIZE   = 20,
    LEGEND_LABEL_GAP  = 5,
    DEGREES_IN_CIRCLE = 360,
    MAX_SLICES        = 6;

  /** Constructor
   *  Receive:       sliceLabels, an array of String;
   *                 sliceValues, an array of double
   *  Precondition:  sliceLabels.length = sliceValues.length
   *                 && sliceLabels.length >= 2
   *                 && sliceLabels.length <= MAX_SLICES
   *                 && sliceValues.length >= 2
   *                 && sliceValues.length <= MAX_SLICES.
   *  Postcondition: for each index i in 0..MAX_SLICES-1:
   *                  mySliceLabel[i].equals(sliceLabels[i])
   *                 && mySliceValue[i] == sliceValues[i]
   *                 && mySlices == sliceValues.length
   *                 && myWidth and myHeight are set appropriately.
   */
  public PieChartPanel(final String [] sliceLabels, double [] sliceValues)
  {
    if (sliceLabels.length != sliceValues.length)
      Controller.fatal("PieChart(String[], double[])",
                        "arrays different lengths");
    else if (sliceLabels.length	< 2)
      Controller.fatal("PieChart(String[], double[])",
                        "2 values minimum");
    else if (sliceLabels.length > MAX_SLICES)
      Controller.fatal("PieChart(String[], double[])",
                        MAX_SLICES + " values maximum");

    mySlices = sliceValues.length;
    mySliceLabel = new String[mySlices];
    mySliceValue = new double[mySlices];

    for (int i = 0; i < mySlices; i++)
    {
      mySliceLabel[i] = sliceLabels[i];
      mySliceValue[i] = sliceValues[i];
    }

    myFormatter = NumberFormat.getPercentInstance();
    mySlicePct = new double[mySlices];
    computePercentages();

    myWidth  = MARGIN + CIRCLE_DIAMETER + MARGIN;
    myHeight = LEGEND_OFFSET + mySlices * LEGEND_BOX_SIZE + MARGIN;
    setSize(myWidth, myHeight);
    setBackground(Color.white);
  }

  /** Utility method to compute percentages for slices.
   *  Postcondition: Attribute variable mySlicePercent has been initialized.
   */
  private void computePercentages()
  {
    double valuesSum = 0.0;

    for (int i = 0; i < mySlices; i++)
      valuesSum += mySliceValue[i];

    for (int i = 0; i < mySlices; i++)
      mySlicePct[i] = mySliceValue[i] / valuesSum;
  }


  /** Method to paint myself
   *  Receive:       pen, a Graphics object
   *  Postcondition: my slices, legend-boxes, and labels
   *                  have been painted on screen.
   */
  public void paintComponent(Graphics pen)
  {
    super.paintComponent(pen);
    setSize(myWidth, myHeight);

    int startAngle = 0;
    int arcAngle = 0;
 
    for (int i = 0; i < mySlices; i++)
    {
      startAngle += arcAngle;
      arcAngle = round(mySlicePct[i] * DEGREES_IN_CIRCLE);
      drawSlice(i, startAngle, arcAngle, pen);
    }
  }

  /** Utility to draw 1 slice, its legend-box, and label.
   *  Receive:       sliceNumber, the index of the slice we are drawing;
   *                  startAngle and arcAngle of the slice, two ints;
   *                  pen, a Graphics object
   *  Precondition:  0 <= sliceNumber && sliceNumber < mySlices
   *                 &&  startAngle and arcAngle are valid angles.
   *  Postcondition: slice sliceNumber has been drawn correctly.
   */
  private void drawSlice(int sliceNumber, int startAngle, int arcAngle,
                          Graphics pen)
  {
    if (sliceNumber < 0 || sliceNumber >= mySlices)
      Controller.fatal("drawSlice()", "invalid sliceNumber");
    else if (startAngle < 0)
      Controller.fatal("drawSlice()", "negative startAngle");
    else if (arcAngle < 0)
      Controller.fatal("drawSlice()", "negative arcAngle");

    pen.setColor(mySliceColor[sliceNumber]);
    pen.fillArc(MARGIN, TOP_MARGIN, CIRCLE_DIAMETER, CIRCLE_DIAMETER,
                 startAngle, arcAngle);
    pen.fillRect(MARGIN,
                 LEGEND_OFFSET + sliceNumber*LEGEND_BOX_SIZE,
                 LEGEND_BOX_SIZE, LEGEND_BOX_SIZE);
     pen.drawString("(" + myFormatter.format(mySlicePct[sliceNumber]) + ") - "
                   + mySliceLabel[sliceNumber],
                   MARGIN + LEGEND_BOX_SIZE + LEGEND_LABEL_GAP,
                   LEGEND_OFFSET + (sliceNumber+1)*LEGEND_BOX_SIZE);
  }

  /** Mutator so that programmers can customize slice colors.
   *  Receive:       int sliceNumber, Color newColor
   *  Precondition:  0 <= sliceNumber && sliceNumber < mySlices
   *                  && newColor is a valid Color.
   *  Postcondition: mySliceColor[sliceNumber] == newColor.
   */
  public void setSliceColor(int sliceNumber, Color newColor)
  {
    if (sliceNumber < 0 || sliceNumber >= mySlices)
      Controller.fatal("setSliceColor(int,Color)", " invalid slice "
                      + sliceNumber);

    mySliceColor[sliceNumber] = newColor;
    repaint();
  }

  /** Mutator so that programmers can modify slice percentages.
   *  Receive        int sliceNumber, double newValue
   *  Precondition:  0 <= sliceNumber && sliceNumber < mySlices
   *                  && newValue is a valid percentage.
   *  Postcondition: mySliceValue[sliceNumber] == newValue.
   */
  public void setSliceValue(int sliceNumber, double newValue)
  {
    if (sliceNumber < 0 || sliceNumber >= mySlices)
      Controller.fatal("setSlicePercent(int,double)", " invalid slice "
                      + sliceNumber);

    mySliceValue[sliceNumber] = newValue;
    computePercentages();
    repaint();
  }

  /** double-to-int rounding utility
   *  Receive: value, a double
   *  Return:  an int == value rounded off
   */
  private int round(double value)
  {
    return (int) (value + 0.5);
  }

 //--- Attribute Variables ---
  private String [] mySliceLabel;   // label for a given slice
  private double [] mySliceValue;   // raw value for a given slice
  private double [] mySlicePct;     // percentage for a given slice
  private Color  [] mySliceColor    // color for a given slice
                      = {Color.green, Color.blue, Color.red,
                         Color.orange, Color.gray, Color.cyan};
  private int      mySlices,        // number of slices
                   myWidth,         // window-frame dimensions
                   myHeight;
  private NumberFormat myFormatter; // display numbers as percentages
}

